###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.format_newline import FormatNewline
from cmake_tidy.formatting.utils.invocation.command_realign_modifier import CommandRealignModifier
from cmake_tidy.formatting.utils.tokens import Tokens
from cmake_tidy.formatting.utils.updaters.keyword_state_updater import KeywordStateUpdater
from cmake_tidy.lexical_data import KeywordVerifier


class CommandSplitter:
    def __init__(self, state: dict, settings: dict):
        self.__prepare_state(state)
        self.__state_updater = KeywordStateUpdater(self.__state, settings)
        self.__verifier = KeywordVerifier(settings)
        self.__settings = settings

    def split(self, invocation: dict) -> list:
        invocation['arguments'] = self.__split_args_to_newlines(invocation['arguments'])
        invocation['arguments'] = self.__realign(invocation)
        return invocation['arguments'] + self.__add_closing_parenthesis_separator(invocation)

    def __realign(self, invocation: dict) -> list:
        return CommandRealignModifier(self.__state, self.__settings).realign(invocation)

    def __split_args_to_newlines(self, args: list) -> list:
        if self.__verifier.is_keyword_or_property(args[0]):
            args = [FormatNewline(self.__state, self.__settings)(1)] + args
        return [self.__handle_argument(arg) for arg in args]

    def __handle_argument(self, arg: str) -> str:
        self.__state_updater.update_state(arg)
        return self.__get_converted_whitespace() if arg == ' ' else arg

    def __add_closing_parenthesis_separator(self, invocation: dict) -> list:
        if self.__settings['closing_parentheses_in_newline_when_split'] and \
                not self.__is_last_element_newline(invocation):
            return [FormatNewline(self.__state, self.__settings)(1)]
        return []

    @staticmethod
    def __is_last_element_newline(invocation: dict) -> bool:
        return Tokens.is_spacing_token(invocation['arguments'][-1]) or \
               Tokens.is_line_comment(invocation['arguments'][-1])

    def __get_converted_whitespace(self) -> str:
        return FormatNewline(self.__state, self.__settings)(1)

    def __prepare_state(self, state: dict) -> None:
        self.__state = state.copy()
        if self.__state['has_first_class_keyword']:
            self.__state['indent'] -= 1
        self.__state['has_first_class_keyword'] = False
        self.__state['keyword_argument'] = False



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


class _Executor:
    def __init__(self, state: dict, name: str, to_be_called):
        self.__state = state
        self.__name = name
        self.__to_be_called = to_be_called

    def __call__(self, arguments=None):
        if arguments is None:
            value = self.__to_be_called()
        else:
            value = self.__to_be_called(arguments)
        self.__state['last'] = self.__name
        return value


class CMakeFormatDispatcher:
    def __init__(self, state: dict):
        self.__state = state
        self.__dict = {}

    def __setitem__(self, key, value):
        if not callable(value):
            raise TypeError('Only callable values accepted')
        self.__dict[key] = _Executor(self.__state, key, value)

    def __getitem__(self, item):
        return self.__dict[item]

    def __contains__(self, item):
        return item in self.__dict



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re

from cmake_tidy.formatting.utils.tokens import Tokens
from cmake_tidy.lexical_data import KeywordVerifier


class CommandInvocationStateUpdater:
    def __init__(self, state: dict):
        self.__state = state

    def update_state(self, function_name: str) -> None:
        self.__update_indent_state(function_name)
        self.__state['keyword_argument'] = False
        self.__state['has_first_class_keyword'] = False

    def __update_indent_state(self, function_name: str) -> None:
        if not self.__is_start_of_special_command(function_name):
            self.__state['indent'] -= 1
        if self.__state['has_first_class_keyword']:
            self.__state['indent'] -= 1
        if self.__is_end_of_special_command(function_name):
            self.__state['indent'] -= 1
        if KeywordVerifier.is_conditional_invocation(function_name):
            self.__state['indent'] -= 1

    def __is_start_of_special_command(self, original: str) -> bool:
        return any([self.__matches(token, original.lower()) for token in Tokens.start_tokens()])

    def __is_end_of_special_command(self, original: str) -> bool:
        return any([self.__matches(token, original.lower()) for token in Tokens.end_tokens()])

    @staticmethod
    def __matches(token: str, data: str) -> bool:
        return re.match(r'^' + re.escape(token) + r'\s?\(', data) is not None



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.invocation.command_formatter import CommandFormatter
from cmake_tidy.formatting.utils.invocation.condition_formatter import ConditionFormatter
from cmake_tidy.lexical_data import KeywordVerifier


def format_command_invocation(state: dict, settings: dict, invocation: dict) -> str:
    if KeywordVerifier.is_conditional_invocation(invocation['function_name']):
        return ConditionFormatter(state, settings).format(invocation)
    return CommandFormatter(state, settings).format(invocation)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import ply.yacc as yacc

from cmake_tidy.parsing.cmake_lexer import CMakeLexer
from cmake_tidy.lexical_data.elements import PrimitiveElement, ComplexElement


class ParsingError(Exception):
    pass


class CMakeParser:
    def __init__(self) -> None:
        self.lexer = CMakeLexer()
        self.tokens = self.lexer.tokens
        self.parser = yacc.yacc(module=self)
        self.__lineno = 1

    @staticmethod
    def p_file(p):
        """file : file file_element
                | file_element
                | empty"""
        if p[1].name == 'file':
            p[0] = p[1].add(p[2])
        else:
            p[0] = ComplexElement('file').add(p[1])

    @staticmethod
    def p_file_element(p):
        """file_element : line_ending
                        | spaces
                        | command_invocation"""
        p[0] = p[1]

    @staticmethod
    def p_command_invocation(p):
        """command_invocation : start_cmd_invoke arguments end_cmd_invoke"""
        p[0] = ComplexElement('command_invocation')
        for element in p[1:]:
            p[0].add(element)

    @staticmethod
    def p_parentheses(p):
        """parentheses : start_parenthesis arguments end_parenthesis"""
        p[0] = ComplexElement('parentheses')
        for element in p[1:]:
                p[0].add(element)

    @staticmethod
    def p_arguments(p):
        """arguments : arguments argument
                     | argument"""
        if p[1].name == 'arguments':
            p[0] = p[1].add(p[2])
        else:
            p[0] = ComplexElement('arguments').add(p[1])

    @staticmethod
    def p_argument(p):
        """argument : bracket_argument
                    | quoted_argument
                    | unquoted_argument
                    | arguments_separation
                    | parentheses
                    | empty"""
        p[0] = p[1]

    @staticmethod
    def p_arguments_separation(p):
        """arguments_separation : spaces
                                | line_ending"""
        p[0] = p[1]

    @staticmethod
    def p_quoted_argument(p):
        """quoted_argument : QUOTED_ARGUMENT_START quoted_argument_content QUOTED_ARGUMENT_END"""
        p[0] = PrimitiveElement('quoted_argument', p[2].values)

    @staticmethod
    def p_bracket_argument(p):
        """bracket_argument : BRACKET_ARGUMENT_START bracket_argument_content BRACKET_ARGUMENT_END"""
        p[0] = ComplexElement('bracket_argument') \
            .add(PrimitiveElement('bracket_start', p[1])) \
            .add(PrimitiveElement('bracket_argument_content', p[2].values)) \
            .add(PrimitiveElement('bracket_end', p[3]))

    @staticmethod
    def p_line_ending(p):
        """line_ending : line_comment newlines
                       | newlines"""
        p[0] = ComplexElement('line_ending')
        for element in p[1:]:
            p[0].add(element)

    @staticmethod
    def p_unquoted_argument(p):
        """unquoted_argument : UNQUOTED_ARGUMENT"""
        p[0] = PrimitiveElement('unquoted_argument', p[1])

    @staticmethod
    def p_line_comment(p):
        """line_comment : LINE_COMMENT"""
        p[0] = PrimitiveElement('line_comment', p[1])

    def p_newlines(self, p):
        """newlines : NEWLINES"""
        newlines = len(p[1])
        self.__lineno += newlines
        p[0] = PrimitiveElement('newlines', newlines)

    @staticmethod
    def p_empty(p):
        """empty :"""
        p[0] = PrimitiveElement()

    @staticmethod
    def p_bracket_argument_content(p):
        """bracket_argument_content : bracket_argument_content bracket_argument_content_element
                                    | bracket_argument_content_element"""
        _create_content(p, 'bracket_argument_content')

    @staticmethod
    def p_quoted_argument_content(p):
        """quoted_argument_content : quoted_argument_content quoted_argument_content_element
                                   | quoted_argument_content_element
                                   | empty"""
        _create_content(p, 'quoted_argument_content')

    @staticmethod
    def p_bracket_argument_content_element(p):
        """bracket_argument_content_element : BRACKET_ARGUMENT_CONTENT"""
        p[0] = _get_content_element(p[1])

    @staticmethod
    def p_quoted_argument_content_element(p):
        """quoted_argument_content_element : QUOTED_ARGUMENT_CONTENT"""
        p[0] = _get_content_element(p[1])

    @staticmethod
    def p_spaces(p):
        """spaces : SPACES"""
        p[0] = PrimitiveElement('spaces', p[1])

    @staticmethod
    def p_start_parenthesis(p):
        """start_parenthesis : BEGIN_PARENTHESIS"""
        p[0] = PrimitiveElement('parenthesis_start', p[1])

    @staticmethod
    def p_end_parenthesis(p):
        """end_parenthesis : END_PARENTHESIS"""
        p[0] = PrimitiveElement('parenthesis_end', p[1])

    @staticmethod
    def p_start_cmd_invoke(p):
        """start_cmd_invoke : COMMAND_INVOCATION_START"""
        p[0] = PrimitiveElement('start_cmd_invoke', p[1])

    @staticmethod
    def p_end_cmd_invoke(p):
        """end_cmd_invoke : COMMAND_INVOCATION_END"""
        p[0] = PrimitiveElement('end_cmd_invoke', p[1])

    def p_error(self, p):
        if p is not None:
            raise ParsingError(f'Illegal symbol in line {self.__lineno}: \"{p.value}\"')
        raise ParsingError('Missing symbol - some invocation is not properly finished')

    def parse(self, data: str):
        return self.parser.parse(data)


def _get_content_element(data) -> PrimitiveElement:
    return PrimitiveElement('content_element', str(data))


def _create_content(p, name: str):
    if p[1].name is name:
        p[0] = p[1]
        p[0].values += p[2].values
    else:
        p[0] = PrimitiveElement(name, p[1].values)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from pathlib import Path

from cmake_tidy.commands.format import FormatConfiguration


class WriterError(Exception):
    pass


def write_to_file(file: Path, data: str):
    with file.open('w') as file:
        file.write(data)


class OutputWriter:
    def __init__(self, configuration: FormatConfiguration):
        self.__config = configuration

    def write(self, data: str) -> None:
        if self.__config.inplace:
            self.__try_writing_to_file(data)
        else:
            print(data)

    def __try_writing_to_file(self, data: str) -> None:
        try:
            write_to_file(self.__config.file, data)
        except PermissionError:
            raise WriterError(f'File {self.__config.file} is read-only!')



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import sys

from cmake_tidy.commands.format.format_command import FormatCommand
from cmake_tidy.utils import ExitCodes
from cmake_tidy.utils.command_line_handling.command_line_parser import CommandLineParser
from cmake_tidy.version import show_version


def run():
    main()


def main(args=sys.argv[1:]):
    parser = __init_parser()
    arguments = parser.parse(args)
    sys.exit(__execute(arguments, parser))


def __init_parser() -> CommandLineParser:
    parser = CommandLineParser()
    parser.add_command(FormatCommand)
    return parser


def __execute(arguments, parser) -> int:
    if arguments.sub_command:
        return arguments.func(arguments)
    __handle_basic_functionality(arguments, parser)
    return ExitCodes.SUCCESS


def __handle_basic_functionality(arguments, parser) -> None:
    if arguments.version:
        show_version()
    else:
        parser.print_help()



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re

from cmake_tidy.lexical_data import KeywordVerifier


class FormatStartCommandInvocation:
    def __init__(self, state: dict, settings: dict):
        self.__state = state
        self.__settings = settings

    def __call__(self, data: str) -> str:
        self.__update_state(data)
        return self.__format_data(data)

    def __update_state(self, data: str):
        if KeywordVerifier.is_conditional_invocation(data):
            self.__state['indent'] += 1
        self.__state['indent'] += 1

    def __format_data(self, original: str) -> str:
        formatted = self.__remove_whitespaces_after_name(original)
        formatted = self.__unify_command_name(formatted)
        return self.__add_spacing_if_needed(formatted)

    def __unify_command_name(self, formatted: str) -> str:
        if self.__settings.get('force_command_lowercase'):
            return formatted.lower()
        return formatted

    def __add_spacing_if_needed(self, formatted: str) -> str:
        if self.__is_spacing_needed(formatted):
            return formatted.replace('(', ' (')
        return formatted

    def __is_spacing_needed(self, formatted: str) -> bool:
        return self.__settings.get('space_between_command_and_begin_parentheses') or \
               self.__should_add_space_for_conditional(formatted)

    def __should_add_space_for_conditional(self, formatted: str) -> bool:
        return self.__settings.get('space_after_loop_condition') and \
               KeywordVerifier.is_conditional_invocation(formatted)

    @staticmethod
    def __remove_whitespaces_after_name(original: str) -> str:
        formatted = re.sub(r'\s+', '', original)
        return formatted



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import ply.lex as lex


class CMakeLexer:
    states = (
        ('commandinvocation', 'exclusive'),
        ('bracketargument', 'exclusive'),
        ('quotedargument', 'exclusive'),
        ('insideparentheses', 'exclusive')
    )
    tokens = ['NEWLINES',
              'LINE_COMMENT',
              'SPACES',
              'COMMAND_INVOCATION_START',
              'COMMAND_INVOCATION_END',
              'BRACKET_ARGUMENT_START',
              'BRACKET_ARGUMENT_CONTENT',
              'BRACKET_ARGUMENT_END',
              'QUOTED_ARGUMENT_START',
              'QUOTED_ARGUMENT_CONTENT',
              'QUOTED_ARGUMENT_END',
              'UNQUOTED_ARGUMENT',
              'BEGIN_PARENTHESIS',
              'END_PARENTHESIS']

    t_INITIAL_commandinvocation_insideparentheses_LINE_COMMENT = r'\#((?=\n)|[^\n]+)'
    t_INITIAL_commandinvocation_insideparentheses_SPACES = r'[ \t]+'
    t_quotedargument_QUOTED_ARGUMENT_CONTENT = r'(?<!\\)\\\"'
    t_commandinvocation_insideparentheses_UNQUOTED_ARGUMENT = r'[^ \t\(\)\#\"\\\n]+'

    def __init__(self) -> None:
        self.bracket_argument_size = -1
        self.lexer = lex.lex(module=self)

    @staticmethod
    def t_begin_commandinvocation(t: lex.Token) -> lex.Token:
        r"""[A-Za-z_][A-Za-z0-9_]*[ \t]*\("""
        t.type = 'COMMAND_INVOCATION_START'
        t.lexer.push_state('commandinvocation')
        return t

    def t_commandinvocation_insideparentheses_begin_insideparentheses(self, t: lex.Token) -> lex.Token:
        r"""\("""
        t.type = 'BEGIN_PARENTHESIS'
        t.lexer.push_state('insideparentheses')
        return t

    def t_commandinvocation_insideparentheses_begin_bracketargument(self, t: lex.Token) -> lex.Token:
        r"""\[=*\["""
        t.type = 'BRACKET_ARGUMENT_START'
        self.bracket_argument_size = len(t.value)
        t.lexer.push_state('bracketargument')
        return t

    def t_commandinvocation_insideparentheses_begin_quotedargument(self, t: lex.Token) -> lex.Token:
        r"""\""""
        t.type = 'QUOTED_ARGUMENT_START'
        t.lexer.push_state('quotedargument')
        return t

    @staticmethod
    def t_commandinvocation_end(t: lex.Token) -> lex.Token:
        r"""\)"""
        return _end_state(t, 'COMMAND_INVOCATION_END')

    @staticmethod
    def t_insideparentheses_end(t: lex.Token) -> lex.Token:
        r"""\)"""
        return _end_state(t, 'END_PARENTHESIS')

    @staticmethod
    def t_quotedargument_end(t: lex.Token) -> lex.Token:
        r"""\""""
        return _end_state(t, 'QUOTED_ARGUMENT_END')

    def t_bracketargument_end(self, t: lex.Token) -> lex.Token:
        r"""\]=*\]"""
        if self.bracket_argument_size == len(t.value):
            t.lexer.pop_state()
            t.type = 'BRACKET_ARGUMENT_END'
        else:
            t.type = 'BRACKET_ARGUMENT_CONTENT'
        return t

    @staticmethod
    def t_INITIAL_commandinvocation_insideparentheses_NEWLINES(t: lex.LexToken) -> lex.LexToken:
        r"""\n+"""
        t.lexer.lineno += len(t.value)
        return t

    @staticmethod
    def t_INITIAL_commandinvocation_insideparentheses_error(t: lex.LexToken) -> lex.LexToken:
        return _skip_one_and_return_with(t, '?')

    @staticmethod
    def t_bracketargument_error(t: lex.LexToken) -> lex.LexToken:
        return _skip_one_and_return_with(t, 'BRACKET_ARGUMENT_CONTENT')

    @staticmethod
    def t_quotedargument_error(t: lex.LexToken) -> lex.LexToken:
        return _skip_one_and_return_with(t, 'QUOTED_ARGUMENT_CONTENT')

    def analyze(self, data: str) -> list:
        self.lexer.input(data)
        data = []

        while True:
            tok = self.lexer.token()
            if not tok:
                break
            data.append(tok)

        return data


def _skip_one_and_return_with(element: lex.LexToken, token_type: str) -> lex.LexToken:
    element.lexer.skip(1)
    element.type = token_type
    element.value = element.value[0]
    return element


def _end_state(element: lex.LexToken, token_type: str) -> lex.LexToken:
    element.lexer.pop_state()
    element.type = token_type
    return element



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re

from cmake_tidy.formatting.utils.tokens import Tokens


class LineLengthCalculator:
    def __init__(self, settings: dict):
        self.__settings = settings

    def calculate(self, invocation: str) -> int:
        invocation = invocation.replace('\t', ' ' * self.__settings['tab_size'])
        invocation = re.sub(Tokens.get_reindent_regex(), '', invocation)
        invocation = re.sub(Tokens.remove_spaces(), '', invocation)
        return len(invocation)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.invocation import format_command_invocation
from cmake_tidy.formatting.utils.updaters.command_invocatin_state_updater import CommandInvocationStateUpdater


class FormatCommandInvocation:
    def __init__(self, state: dict, settings: dict):
        self.__state = state
        self.__state_handler = CommandInvocationStateUpdater(state)
        self.__settings = settings

    def __call__(self, data: list) -> str:
        command_invocation = self.__prepare_data(data)
        formatted = format_command_invocation(self.__state, self.__settings, command_invocation)
        self.__state_handler.update_state(command_invocation['function_name'])
        return formatted

    @staticmethod
    def __prepare_data(data: list) -> dict:
        return {'function_name': data[0],
                'arguments': data[1] if len(data) == 3 else [],
                'closing': data[2] if len(data) == 3 else data[1]}



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import inspect


class ConfigurationError(Exception):
    pass


class Configuration:
    def __init__(self, arguments):
        self._config = {}
        for name in self.__get_all_property_names():
            if arguments.get(name) is not None:
                self._config[name] = arguments.get(name)

    @property
    def all_properties(self):
        return {name: getattr(self, name) for name in self.__get_all_property_names()}

    @staticmethod
    def _property_name():
        return inspect.stack()[1][3]

    def __get_all_property_names(self) -> list:
        properties = [name for name, func in inspect.getmembers(self.__class__, self.__is_property)]
        properties.remove('all_properties')
        return properties

    @staticmethod
    def __is_property(v) -> bool:
        return isinstance(v, property)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import json
import re
from pathlib import Path

from cmake_tidy.formatting.utils.tokens import Tokens


class KeywordVerifier:
    __FIRST_CLASS_KEYWORDS = ['PROPERTIES', 'PROPERTY']
    __PROPERTIES = dict()

    @staticmethod
    def __init_properties():
        if not KeywordVerifier.__PROPERTIES:
            with (Path(__file__).parent / 'keyword_list.json').open() as file:
                KeywordVerifier.__PROPERTIES = json.load(file)

    def __init__(self, settings: dict):
        KeywordVerifier.__init_properties()
        self.__settings = settings

    @staticmethod
    def is_first_class_keyword(data: str) -> bool:
        data = data.replace(Tokens.reindent(1), '')
        return data in KeywordVerifier.__FIRST_CLASS_KEYWORDS

    def get_cmake_properties_version(self) -> str:
        return self.__PROPERTIES["cmake_version"]

    def is_keyword_or_property(self, data: str) -> bool:
        return self.is_property(data) or self.is_keyword(data)

    def is_keyword(self, data: str) -> bool:
        data = data.replace(Tokens.reindent(1), '')
        return self.__is_one_of_defined_keywords(data) or \
               self.__should_be_handled_as_keyword(data) or \
               self.is_first_class_keyword(data) or \
               self.__is_keyword_in_cmake(data)

    @staticmethod
    def is_conditional_invocation(data: str) -> bool:
        data = data.lower().replace(' ', '')
        return any([token == data[:-1] for token in Tokens.conditional_tokens()]) and data[-1] == '('

    @staticmethod
    def is_double_keyword(first: str, second: str) -> bool:
        first = first.replace(Tokens.reindent(1), '')
        second = second.replace(Tokens.reindent(1), '')
        return any(keyword.startswith(first) and keyword.endswith(second) \
                   for keyword in KeywordVerifier.__PROPERTIES['double-keywords'])

    @staticmethod
    def is_command_keyword(data: str) -> bool:
        data = data.replace(Tokens.reindent(1), '')
        return data == 'COMMAND' or data == 'ARGS'

    def __is_one_of_defined_keywords(self, data: str) -> bool:
        return self.__settings.get('keywords') and data in self.__settings['keywords']

    def __should_be_handled_as_keyword(self, data: str) -> bool:
        upper_case_regex = r'^([A-Z]+_?)+[A-Z]$'
        return self.__settings.get('unquoted_uppercase_as_keyword') and re.match(upper_case_regex, data)

    def is_property(self, data: str) -> bool:
        data = data.replace(Tokens.reindent(1), '')
        return data in KeywordVerifier.__PROPERTIES["properties_full_names"] or \
               self.__is_property_regex_starting(data) or \
               self.__is_property_ending_with(data)

    @staticmethod
    def __is_keyword_in_cmake(data: str) -> bool:
        return any([data in KeywordVerifier.__PROPERTIES["keywords"]])

    @staticmethod
    def __is_property_ending_with(data: str) -> bool:
        return any([data.endswith(token) for token in KeywordVerifier.__PROPERTIES['properties_ending_with']])

    @staticmethod
    def __is_property_regex_starting(data: str) -> bool:
        return any([data.startswith(token) for token in KeywordVerifier.__PROPERTIES['properties_starting_with']])



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from typing import List, Tuple

from cmake_tidy.formatting.utils.tokens import Tokens
from cmake_tidy.lexical_data import KeywordVerifier


class CommandRealignModifier:
    __DIFF_BETWEEN_KEYWORD_AND_VALUE = 2

    def __init__(self, state: dict, settings: dict):
        self.__state = state
        self.__verifier = KeywordVerifier(settings)
        self.__settings = settings

    def realign(self, invocation: dict) -> list:
        invocation['arguments'] = self.__realign_properties_if_needed(invocation['arguments'])
        invocation['arguments'] = self.__realign_double_keywords(invocation['arguments'])
        invocation['arguments'] = self.__realign_get_property(invocation)
        invocation['arguments'] = self.__realign_set_property(invocation)
        invocation['arguments'] = self.__realign_commands_if_needed(invocation['arguments'])
        return self.__realign_keyword_values_if_needed(invocation['arguments'])

    def __realign_commands_if_needed(self, args: List[str]) -> list:
        return self.__realign_commands(args) if self.__settings.get('keep_command_in_single_line') else args

    def __realign_commands(self, args: List[str]) -> list:
        diff = 0 if self.__settings.get('keyword_and_single_value_in_one_line') else CommandRealignModifier.__DIFF_BETWEEN_KEYWORD_AND_VALUE
        for i in range(len(args)):
            if KeywordVerifier.is_command_keyword(args[i]):
                for j in range(i + diff, len(args) - 1):
                    if self.__verifier.is_keyword(args[j + 1]):
                        break
                    if Tokens.is_spacing_token(args[j]):
                        args[j] = ' '
        return args

    def __realign_properties_if_needed(self, args: List[str]) -> list:
        return self.__realign_properties(args) if self.__should_realign_properties() else args

    def __should_realign_properties(self) -> bool:
        return self.__settings['keep_property_and_value_in_one_line'] and self.__state['has_first_class_keyword']

    def __realign_properties(self, args: List[str]) -> list:
        for i in range(len(args) - 2):
            if self.__is_property(args[i]) and self.__should_realign_value_after_keyword(args, i):
                args[i + 1] = ' '
        return args

    def __is_property(self, argument: str) -> bool:
        return not self.__verifier.is_first_class_keyword(argument) and self.__verifier.is_keyword_or_property(argument)

    def __realign_double_keywords(self, args: List[str]) -> list:
        for i in range(len(args) - self.__DIFF_BETWEEN_KEYWORD_AND_VALUE):
            if self.__verifier.is_double_keyword(args[i], args[i + self.__DIFF_BETWEEN_KEYWORD_AND_VALUE]):
                args[i + 1] = ' '
        return args

    def __realign_get_property(self, invocation: dict) -> list:
        if invocation['function_name'].startswith('get_property'):
            return self.__replace_newline_with_space_after_property_keyword(invocation['arguments'])
        return invocation['arguments']

    def __realign_set_property(self, invocation: dict) -> list:
        if invocation['function_name'].startswith('set_property'):
            return self.__realign_property_in_set_function(invocation['arguments'])
        return invocation['arguments']

    def __realign_property_in_set_function(self, args: List[str]) -> list:
        for i in range(len(args)):
            if KeywordVerifier.is_first_class_keyword(args[i]):
                args, position = self.__reindent_property_name(args, i)
                if self.__is_possible_to_realign_keyword_values(args) and self.__is_value_realignable(args, position):
                    args[position + 1] = ' '
        return args

    def __is_value_realignable(self, args, position):
        return not Tokens.is_line_comment(args[position + 2]) and \
               self.__get_number_of_arguments(args, position) == 1

    def __get_number_of_arguments(self, args: List[str], start: int) -> int:
        return sum([self.__is_argument(data) for data in args[start + 1:]])

    def __reindent_property_name(self, args: list, start_index: int) -> Tuple[list, int]:
        for i in range(start_index + 1, len(args)):
            if self.__is_argument(args[i]):
                if not args[i].startswith(Tokens.reindent(1)):
                    args[i] = Tokens.reindent(1) + args[i]
                return args, i
        return args, 0

    @staticmethod
    def __is_argument(data: str) -> bool:
        return not (Tokens.is_line_comment(data) or Tokens.is_spacing_token(data))

    def __replace_newline_with_space_after_property_keyword(self, args: List[str]) -> list:
        for i in range(len(args) - CommandRealignModifier.__DIFF_BETWEEN_KEYWORD_AND_VALUE):
            if self.__is_property_followed_by_name(args, i):
                args[i + 1] = ' '
        return args

    def __realign_keyword_values_if_needed(self, args: List[str]) -> list:
        return self.__realign_keyword_values(args) if self.__is_possible_to_realign_keyword_values(args) else args

    def __is_possible_to_realign_keyword_values(self, args: List[str]) -> bool:
        return self.__settings['keyword_and_single_value_in_one_line'] and \
               len(args) > CommandRealignModifier.__DIFF_BETWEEN_KEYWORD_AND_VALUE

    def __realign_keyword_values(self, args: List[str]) -> list:
        for i in range(len(args) - CommandRealignModifier.__DIFF_BETWEEN_KEYWORD_AND_VALUE):
            if self.__should_realign_value_after_keyword(args, i):
                args[i + 1] = ' '
        return args

    def __should_realign_value_after_keyword(self, args: List[str], current_index: int) -> bool:
        return self.__verifier.is_keyword_or_property(args[current_index]) and \
               Tokens.is_spacing_token(args[current_index + 1]) and \
               not self.__verifier.is_keyword_or_property(args[current_index + 2]) and \
               self.__is_single_value(args, current_index)

    def __is_single_value(self, args: List[str], current_index: int) -> bool:
        tokens_diff = 4  # keyword, space, value, space, keyword
        return current_index + tokens_diff >= len(args) or \
               self.__verifier.is_keyword_or_property(args[current_index + tokens_diff])

    @staticmethod
    def __is_property_followed_by_name(args: list, i: int) -> bool:
        return KeywordVerifier.is_first_class_keyword(args[i]) and not Tokens.is_line_comment(args[i + 2])



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re
from abc import ABC, abstractmethod
from typing import List

from cmake_tidy.formatting.utils.invocation.invocation_wrapper import InvocationWrapper
from cmake_tidy.formatting.utils.invocation.line_comments_formatter import LineCommentsFormatter
from cmake_tidy.formatting.utils.line_length_calculator import LineLengthCalculator
from cmake_tidy.formatting.utils.single_indent import get_single_indent
from cmake_tidy.formatting.utils.tokens import Tokens


class InvocationFormatter(ABC):
    def __init__(self, state: dict, settings: dict):
        self._settings = settings
        self._state = state

    @abstractmethod
    def format(self, invocation: dict) -> str:
        pass

    def _is_fitting_in_line(self, command_invocation: dict) -> bool:
        return self._invocation_length(command_invocation) < self._settings['line_length']

    def _newline_indent(self) -> str:
        indent = max(self._state['indent'] - 1, 0)
        return indent * get_single_indent(self._settings)

    def _invocation_length(self, command_invocation: dict) -> int:
        invoke = self._join_command_invocation(command_invocation) + self._newline_indent()
        return LineLengthCalculator(self._settings).calculate(invoke)

    def _join_command_invocation(self, invocation: dict) -> str:
        formatted = invocation['function_name'] + ''.join(invocation['arguments']) + invocation['closing']
        return self.__add_reindent_tokens_where_needed(formatted)

    def _prepare_arguments(self, invocation: dict) -> list:
        invocation['arguments'] = self.__remove_empty_arguments(invocation)
        invocation['arguments'] = self.__remove_whitespace_at_end_of_line(invocation['arguments'])
        invocation['arguments'] = LineCommentsFormatter(self._state, self._settings).format(invocation['arguments'])
        if self.__is_wrappable(invocation):
            invocation['arguments'] = self.__wrap_arguments_if_possible(invocation)
        return invocation['arguments']

    @staticmethod
    def __add_reindent_tokens_where_needed(data: str) -> str:
        data_lower = data.lower()
        if any(data_lower.startswith(token) for token in Tokens.reindent_commands_tokens()):
            return Tokens.reindent(1) + data
        return data

    def __is_wrappable(self, invocation: dict) -> bool:
        return len(invocation['arguments']) > 0 and self._settings['wrap_short_invocations_to_single_line'] is True

    def __wrap_arguments_if_possible(self, invocation: dict) -> list:
        command_invocation = InvocationWrapper().wrap(invocation)
        if self._is_fitting_in_line(command_invocation):
            return command_invocation['arguments']
        else:
            return invocation['arguments']

    @staticmethod
    def __remove_empty_arguments(invocation: dict) -> list:
        return list(filter(len, invocation['arguments']))

    @staticmethod
    def __remove_whitespace_at_end_of_line(args: List[str]) -> list:
        filtered_arguments = []
        for i in range(len(args) - 1):
            if not (Tokens.is_spacing_token(args[i]) and Tokens.is_spacing_token(args[i + 1])):
                filtered_arguments.append(args[i])
        return filtered_arguments + [args[-1]] if args else []



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.format_arguments import FormatArguments
from cmake_tidy.formatting.utils.format_command_invocation import FormatCommandInvocation
from cmake_tidy.formatting.utils.format_end_command_invocation import FormatEndCommandInvocation
from cmake_tidy.formatting.utils.format_file import FormatFile
from cmake_tidy.formatting.utils.format_newline import FormatNewline
from cmake_tidy.formatting.utils.format_spaces import FormatSpaces
from cmake_tidy.formatting.utils.format_start_command_invocation import FormatStartCommandInvocation
from cmake_tidy.formatting.utils.format_unquoted_argument import FormatUnquotedArgument
from cmake_tidy.formatting.utils.format_bracket_argument import FormatBracketArgument



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from pathlib import Path
from typing import Optional

from cmake_tidy.formatting.cmake_formatter import CMakeFormatter
from cmake_tidy.formatting.settings_reader import SettingsReader


def try_read_settings(filepath: Optional[Path]) -> dict:
    return SettingsReader().try_loading_format_settings(filepath)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.tokens import Tokens


class InvocationWrapper:
    def wrap(self, invocation: dict) -> dict:
        wrapped_invoke = self.__prepare_wrapped_invocation(invocation)
        wrapped_invoke['arguments'] = [e if not Tokens.is_spacing_token(e) else ' ' for e in
                                       wrapped_invoke['arguments']]
        return wrapped_invoke

    @staticmethod
    def __prepare_wrapped_invocation(invocation: dict) -> dict:
        new_invoke = invocation.copy()
        if Tokens.is_spacing_token(new_invoke['arguments'][0]):
            new_invoke['arguments'] = new_invoke['arguments'][1:]
        if Tokens.is_spacing_token(new_invoke['arguments'][-1]):
            new_invoke['arguments'] = new_invoke['arguments'][:-1]
        return new_invoke



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import sys
from abc import ABC, abstractmethod

from cmake_tidy.utils import ExitCodes


class Command(ABC):
    def __init__(self, parser, command_name: str, description: str):
        self._command_parser = parser.add_parser(name=command_name, help=description)
        self._command_parser.set_defaults(func=self.execute_command)
        self._command_name = command_name

    @abstractmethod
    def execute_command(self, args) -> int:
        pass

    def _handle_error(self, raised_error: Exception) -> int:
        print(f'cmake-tidy {self._command_name}: {str(raised_error)}', file=sys.stderr)
        return ExitCodes.FAILURE



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re

from cmake_tidy.formatting.utils.single_indent import get_single_indent
from cmake_tidy.formatting.utils.tokens import Tokens


class FormatFile:
    def __init__(self, settings: dict):
        self.__settings = settings

    def __call__(self, data: list) -> str:
        processed = self.__cleanup_end_invocations(''.join(data))
        processed = self.__cleanup_whitespaces_at_line_ends(processed)
        return self.__cleanup_reindent_all(processed)

    def __cleanup_end_invocations(self, formatted_file: str) -> str:
        for pattern in Tokens.get_reindent_patterns_list(3, get_single_indent(self.__settings)):
            formatted_file = re.sub(pattern, '', formatted_file)
        return formatted_file

    @staticmethod
    def __cleanup_whitespaces_at_line_ends(processed: str) -> str:
        return re.sub('[ \t]*' + Tokens.remove_spaces(), '', processed)

    @staticmethod
    def __cleanup_reindent_all(data: str) -> str:
        return re.sub('[ \t]*' + Tokens.reindent(), '', data)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting import utils
from cmake_tidy.formatting.cmake_format_dispatcher import CMakeFormatDispatcher
from cmake_tidy.lexical_data.elements import Element
from cmake_tidy.utils.proxy_visitor import ProxyVisitor


class CMakeFormatter:
    __settings: dict

    def __init__(self, format_settings: dict):
        self.__state = {'indent': 0, 'last': None, 'keyword_argument': False, 'has_first_class_keyword': False}
        self.__settings = format_settings

        self.__formatters = CMakeFormatDispatcher(self.__state)
        self.__formatters['newlines'] = utils.FormatNewline(self.__state, self.__settings)
        self.__formatters['unquoted_argument'] = utils.FormatUnquotedArgument(self.__state, self.__settings)
        self.__formatters['start_cmd_invoke'] = utils.FormatStartCommandInvocation(self.__state, self.__settings)
        self.__formatters['command_invocation'] = utils.FormatCommandInvocation(self.__state, self.__settings)
        self.__formatters['file'] = utils.FormatFile(self.__settings)
        self.__formatters['spaces'] = utils.FormatSpaces(self.__settings, self.__state)
        self.__formatters['arguments'] = utils.FormatArguments(self.__state)
        self.__formatters['end_cmd_invoke'] = utils.FormatEndCommandInvocation(self.__state)
        self.__formatters['bracket_argument'] = utils.FormatBracketArgument()
        self.__formatters['line_comment'] = lambda data: data
        self.__formatters['bracket_start'] = lambda data: data
        self.__formatters['bracket_end'] = lambda data: data
        self.__formatters['parenthesis_start'] = lambda data: data
        self.__formatters['parenthesis_end'] = lambda data: data
        self.__formatters['bracket_argument_content'] = lambda data: data
        self.__formatters['line_ending'] = lambda data: ''.join(data)
        self.__formatters['quoted_argument'] = lambda data: f'"{data}"'
        self.__formatters['parentheses'] = lambda data: data

    def format(self, data: Element) -> str:
        visitor = ProxyVisitor(self.__formatters)
        formatted_elements = data.accept(visitor)
        return ''.join(formatted_elements)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import json
from json import JSONDecodeError
from pathlib import Path
from typing import Optional

from jsonschema import Draft6Validator, SchemaError, ValidationError


class InvalidSchemaError(Exception):
    pass


class SchemaValidationError(ValidationError):
    pass


class SettingsReader:
    def __init__(self):
        self.__schema = self.__try_reading_schema()
        self.__settings = self.get_default_format_settings()

    def try_loading_format_settings(self, filepath: Optional[Path]) -> dict:
        self.__settings.update(self.__try_reading_settings(filepath))
        return self.__settings

    def __try_reading_settings(self, filepath: Optional[Path]) -> dict:
        try:
            user_define_settings = self._read_settings(filepath)
            self.__schema.validate(user_define_settings)
            return user_define_settings
        except (ValidationError, JSONDecodeError) as error:
            raise SchemaValidationError(str(error))

    def __try_reading_schema(self) -> Draft6Validator:
        try:
            schema = self._read_schema_file()
            Draft6Validator.check_schema(schema)
            return Draft6Validator(schema)
        except (FileNotFoundError, OSError, SchemaError, JSONDecodeError):
            raise InvalidSchemaError('JSON schema validation error - please raise issue on github!')

    @staticmethod
    def _read_schema_file() -> dict:
        schema_file = Path(__file__).parent / 'settings.json'
        with schema_file.open() as file:
            return json.load(file)

    def _read_settings(self, filepath: Optional[Path]) -> dict:
        if filepath:
            return self.__try_reading_settings_recursively(filepath)
        else:
            return self.__read_settings_from_filepath(Path.cwd())

    def __try_reading_settings_recursively(self, filepath: Path) -> dict:
        content = dict()
        for path in filepath.parents:
            content = self.__read_settings_from_filepath(path)
            if content:
                break
        return content

    def __read_settings_from_filepath(self, filepath: Path) -> dict:
        self.__settings_file = filepath / '.cmake-tidy.json'
        if self.__settings_file.exists() and self.__settings_file.stat().st_size > 0:
            with self.__settings_file.open() as file:
                return json.load(file)
        return dict()

    @staticmethod
    def get_default_format_settings() -> dict:
        settings = dict()
        settings['succeeding_newlines'] = 2
        settings['tabs_as_spaces'] = False
        settings['tab_size'] = 4
        settings['force_command_lowercase'] = True
        settings['space_between_command_and_begin_parentheses'] = False
        settings['line_length'] = 100
        settings['wrap_short_invocations_to_single_line'] = True
        settings['closing_parentheses_in_newline_when_split'] = True
        settings['unquoted_uppercase_as_keyword'] = False
        settings['space_after_loop_condition'] = False
        settings['keep_property_and_value_in_one_line'] = True
        settings['keyword_and_single_value_in_one_line'] = True
        settings['keep_command_in_single_line'] = True
        settings['condition_splitting_move_and_or_to_newline'] = True
        settings['keywords'] = []
        return settings



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.tokens import Tokens


class FormatEndCommandInvocation:
    def __init__(self, state: dict):
        self.__state = state

    def __call__(self, data: str) -> str:
        self.__update_state()
        return self.__format(data)

    def __update_state(self):
        if self.__state['keyword_argument']:
            self.__state['indent'] -= 1

    def __format(self, data):
        if self.__state['has_first_class_keyword']:
            return Tokens.reindent(3) + data
        if self.__state['keyword_argument']:
            return Tokens.reindent(2) + data
        return Tokens.reindent(1) + data



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.single_indent import get_single_indent


class FormatSpaces:
    def __init__(self, settings: dict, state: dict):
        self.__settings = settings
        self.__state = state

    def __call__(self, data: str) -> str:
        if self.__state['last'] == 'line_ending':
            return ''
        elif self.__state['last'] == 'command_invocation':
            return ' '
        return data.replace('\t', get_single_indent(self.__settings))



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.commands.format.format_configuration import FormatConfiguration
from cmake_tidy.commands.format.output_writer import OutputWriter


def try_create_configuration(arguments) -> FormatConfiguration:
    return FormatConfiguration(arguments=dict(vars(arguments)))



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re


class Tokens:
    @staticmethod
    def start_tokens() -> list:
        return ['macro', 'while', 'foreach', 'if', 'function']

    @staticmethod
    def end_tokens() -> list:
        return [f'end{token}' for token in Tokens.start_tokens()]

    @staticmethod
    def reindent_commands_tokens() -> list:
        return ['elseif', 'else'] + Tokens.end_tokens()

    @staticmethod
    def conditional_tokens() -> list:
        return ['if', 'while', 'foreach', 'elseif']

    @staticmethod
    def reindent(count=99) -> str:
        return f'<cmake-tidy-reindent{count}>'

    @staticmethod
    def remove_spaces() -> str:
        return '<cmake-tidy-remove-space>'

    @staticmethod
    def is_spacing_token(data: str) -> bool:
        data = data.replace(Tokens.remove_spaces(), '')
        return re.match(r'^\s+$', data) is not None

    @staticmethod
    def get_reindent_regex() -> str:
        return r'<cmake-tidy-reindent[0-9]+>'

    @staticmethod
    def get_reindent_patterns_list(count: int, indent: str) -> list:
        return [f'({indent}){{0,{times}}}{Tokens.reindent(times)}' for times in range(1, count + 1)]

    @staticmethod
    def is_line_comment(data: str) -> bool:
        data = re.sub(Tokens.get_reindent_regex(), '', data)
        data = re.sub(Tokens.remove_spaces(), '', data)
        return data.strip().startswith('#')



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re
from iteration_utilities import deepflatten


class FormatArguments:
    __spacing = r'^[ \t]+$'

    def __init__(self, state: dict):
        self.__state = state

    def __call__(self, data: list) -> list:
        return self.__format_arguments(data) if data[0] else []

    def __format_arguments(self, data: list) -> list:
        data = list(deepflatten(data, types=list))
        data = self.__replace_spacings_between_arguments_with_single_space(data)
        data = self.__remove_spacing_from_first_element(data)
        data = self.__remove_spacing_from_last_element(data)
        return data

    @staticmethod
    def __replace_spacings_between_arguments_with_single_space(data: list) -> list:
        return [re.sub(FormatArguments.__spacing, ' ', element) for element in data]

    @staticmethod
    def __remove_spacing_from_first_element(data: list) -> list:
        return data[1:] if re.match(FormatArguments.__spacing, data[0]) else data

    @staticmethod
    def __remove_spacing_from_last_element(data: list) -> list:
        return data[:-1] if re.match(FormatArguments.__spacing, data[-1]) else data



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import re
from typing import List

from cmake_tidy.formatting.utils.tokens import Tokens
from cmake_tidy.lexical_data import KeywordVerifier


class LineCommentsFormatter:
    def __init__(self, state: dict, settings: dict):
        self.__state = state
        self.__settings = settings

    def format(self, args: List[str]) -> list:
        args = self.__reindent_line_comments(args)
        return self.__merge_line_comments_with_whitespaces_before(args)

    def __reindent_line_comments(self, args: List[str]) -> list:
        args = self.__reindent_line_comments_after_keyword(args)
        return self.__reindent_line_comments_at_end_of_invocation(args)

    def __reindent_line_comments_at_end_of_invocation(self, args: List[str]) -> list:
        if self.__state['keyword_argument']:
            self.__try_reindent_all_previous_comments(args, len(args))
        return args

    def __reindent_line_comments_after_keyword(self, args: List[str]) -> list:
        verifier = KeywordVerifier(self.__settings)
        for i in reversed(range(len(args))):
            if verifier.is_keyword(args[i]) and re.match(Tokens.get_reindent_regex(), args[i]):
                self.__try_reindent_all_previous_comments(args, i)
        return args

    def __try_reindent_all_previous_comments(self, args, i):
        try:
            self.__reindent_all_previous_comments(args, i)
        except IndexError:
            pass

    @staticmethod
    def __reindent_all_previous_comments(args: list, start: int) -> None:
        for i in reversed(range(start)):
            if Tokens.is_line_comment(args[i]):
                args[i] = Tokens.reindent(1) + args[i]
            else:
                break

    @staticmethod
    def __merge_line_comments_with_whitespaces_before(args: List[str]) -> list:
        merged = []
        for i in range(len(args) - 1):
            if Tokens.is_spacing_token(args[i]) and Tokens.is_line_comment(args[i + 1]):
                args[i + 1] = args[i] + args[i + 1]
            else:
                merged.append(args[i])
        return merged + [args[-1]] if args else []



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import json
from pathlib import Path

from cmake_tidy.commands import Command
from cmake_tidy.commands.format import try_create_configuration, FormatConfiguration, OutputWriter
from cmake_tidy.formatting import try_read_settings, CMakeFormatter
from cmake_tidy.parsing import CMakeParser
from cmake_tidy.utils.command_line_handling import arguments
from cmake_tidy.utils import ExitCodes
from cmake_tidy.utils.diff import get_unified_diff


class FormatCommand(Command):
    __DESCRIPTION = 'format file to align it to standard'

    def __init__(self, parser):
        super().__init__(parser, 'format', FormatCommand.__DESCRIPTION)

        arguments.dump_config(self._command_parser)
        arguments.inplace(self._command_parser)
        arguments.input_data(self._command_parser)
        arguments.diff(self._command_parser)
        arguments.verbose(self._command_parser)

    def execute_command(self, args) -> int:
        try:
            self.__try_execute_command(args)
            return ExitCodes.SUCCESS
        except Exception as error:
            return self._handle_error(error)

    def __try_execute_command(self, args):
        if args.dump_config:
            self.__try_dump_config(args)
        else:
            config = try_create_configuration(args)
            self.__print_filename_if_needed(args, config)
            formatted_file = self.__try_format_file(config)
            if args.diff:
                print(get_unified_diff(config.input, formatted_file, config.file))
            else:
                self.__try_output_formatted_file(config, formatted_file)

    @staticmethod
    def __print_filename_if_needed(args, config):
        if args.verbose:
            print(f'Formatting file: {config.file}')

    @staticmethod
    def __try_dump_config(args):
        filepath = Path(args.input) if args.input else None
        print(json.dumps(try_read_settings(filepath), indent=2))

    @staticmethod
    def __try_format_file(configuration: FormatConfiguration) -> str:
        parsed_input = CMakeParser().parse(configuration.input)
        format_settings = try_read_settings(configuration.file)
        return CMakeFormatter(format_settings).format(parsed_input)

    @staticmethod
    def __try_output_formatted_file(configuration: FormatConfiguration, formatted_file: str) -> None:
        OutputWriter(configuration).write(formatted_file)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import difflib
from pathlib import Path


def get_unified_diff(original: str, modified: str, filename: Path) -> str:
    diff = difflib.unified_diff(a=original.splitlines(keepends=True),
                                b=modified.splitlines(keepends=True),
                                fromfile=str(filename),
                                tofile=str(filename))
    return ''.join(diff)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import argparse


class CommandLineParser:
    def __init__(self):
        self.__parser = argparse.ArgumentParser(prog='cmake-tidy',
                                                formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        self.__parser.add_argument('-v', '--version', action='store_true', default=False)

        self.__sub_parser = self.__parser.add_subparsers(title='sub-commands',
                                                         dest='sub_command',
                                                         help='see "cmake-tidy <command> --help" to read more '
                                                              'about a specific sub-command.')
        self.__commands = []

    def add_command(self, class_name):
        self.__commands.append(class_name(self.__sub_parser))

    def parse(self, args=None):
        return self.__parser.parse_args(args)

    def print_help(self):
        self.__parser.print_help()



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.updaters.keyword_state_updater import KeywordStateUpdater
from cmake_tidy.formatting.utils.tokens import Tokens
from cmake_tidy.lexical_data import KeywordVerifier


class FormatUnquotedArgument:
    def __init__(self, state: dict, settings: dict):
        self.__verifier = KeywordVerifier(settings)
        self.__state_updater = KeywordStateUpdater(state, settings)
        self.__state = state
        self.__keyword_argument_already_found = False

    def __call__(self, data: str) -> str:
        self.__keyword_argument_already_found = self.__state['keyword_argument']
        self.__state_updater.update_state(data)
        return self.__format_data(data)

    def __format_data(self, data: str) -> str:
        if self.__keyword_argument_already_found and self.__is_reindent_needed(data):
            return Tokens.reindent(1) + data
        return data

    def __is_reindent_needed(self, data):
        return self.__verifier.is_keyword(data) or self.__verifier.is_property(data)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.single_indent import get_single_indent
from cmake_tidy.formatting.utils.tokens import Tokens


class FormatNewline:
    def __init__(self, state: dict, settings: dict):
        self.__state = state
        self.__settings = settings

    def __call__(self, data) -> str:
        return self.__format_newlines(data) + self.__prepare_initial_newline_indent()

    def __prepare_initial_newline_indent(self) -> str:
        return self.__state['indent'] * get_single_indent(self.__settings)

    def __format_newlines(self, number_of_newlines: int) -> str:
        return Tokens.remove_spaces() + '\n' * min(self.__settings['succeeding_newlines'], number_of_newlines)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.lexical_data import KeywordVerifier


class KeywordStateUpdater:
    def __init__(self, state: dict, settings: dict):
        self.__state = state
        self.__verifier = KeywordVerifier(settings)

    def update_state(self, argument: str) -> None:
        if self.__verifier.is_first_class_keyword(argument):
            self.__state['has_first_class_keyword'] = True
            self.__state['indent'] += 1
        elif self.__verifier.is_keyword(argument) or self.__should_indent_property(argument):
            if not self.__state['keyword_argument']:
                self.__state['indent'] += 1
            self.__state['keyword_argument'] = True

    def __should_indent_property(self, argument) -> bool:
        return self.__state['has_first_class_keyword'] and self.__verifier.is_property(argument)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.tokens import Tokens


class FormatBracketArgument:
    def __call__(self, data: list) -> str:
        formatted = ''.join(data)
        if '\n' in data[1]:
            formatted = Tokens.reindent(99) + formatted
        return formatted



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


def input_data(parser):
    parser.add_argument('input',
                        type=str,
                        nargs='?',
                        help='CMake file to be formatted')


def dump_config(parser):
    parser.add_argument('--dump-config',
                        action='store_true',
                        help='Dump to stdout current settings. Script tries to read settings from `.cmake-tidy.json` '
                             'or provides default settings. Precedence of searching `.cmake-tidy.json` is described '
                             'on github')


def inplace(parser):
    parser.add_argument('-i', '--inplace',
                        action='store_true',
                        help='Inplace edit specified <input_data> file')


def diff(parser):
    parser.add_argument('--diff',
                        action='store_true',
                        help='Print to stdout unified diff between original file and formatted version.')


def verbose(parser):
    parser.add_argument('--verbose',
                        action='store_true',
                        help='Print to stdout information about formatted file')



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from cmake_tidy.formatting.utils.invocation.invocation_formatter import InvocationFormatter
from cmake_tidy.formatting.utils.invocation.command_splitter import CommandSplitter


class CommandFormatter(InvocationFormatter):
    def __init__(self, state: dict, settings: dict):
        super().__init__(state, settings)

    def format(self, invocation: dict) -> str:
        invocation['arguments'] = self._prepare_arguments(invocation)
        if not self._is_fitting_in_line(invocation):
            invocation['arguments'] = self.__split_command_to_newlines(invocation)
        return self._join_command_invocation(invocation)

    def __split_command_to_newlines(self, invocation: dict) -> list:
        return CommandSplitter(self._state, self._settings).split(invocation)



import os
import re

from pip.req import parse_requirements
from setuptools import setup, find_packages

install_requirements = parse_requirements('./requirements.txt', session=False)
install_requirements = [str(ir.req) for ir in install_requirements]


def load_version():
    filename = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "cmake_tidy", "version.py"))
    with open(filename, "rt") as version_file:
        return re.search(r"VERSION = '([0-9a-z.-]+)'", version_file.read()).group(1)


setup(
    name='cmake-tidy',
    version=load_version(),
    python_requires='>=3.6.0',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    url='https://github.com/MaciejPatro/cmake-tidy',
    license='LICENSE',
    author='Maciej Patro',
    author_email='maciejpatro@gmail.com',
    install_requires=install_requirements,
    description='cmake-tidy is a tool to format/analyze cmake source files.',
    long_description='For More information visit https://github.com/MaciejPatro/cmake-tidy',
    long_description_content_type='text/plain',
    package_data={
        '': ['*.json', '*.adoc', '*.txt']
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    keywords=['cmake', 'format', 'static-analysis', 'developer', 'tool'],
    entry_points={
        'console_scripts': [
            'cmake-tidy=cmake_tidy.run:run',
        ],
    },
)



from cmake_tidy.lexical_data import KeywordVerifier

VERSION = '0.5.0'


def show_version():
    print(f'cmake-tidy: {VERSION} | supported CMake: {KeywordVerifier(dict()).get_cmake_properties_version()}')



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from abc import ABC, abstractmethod


class Element(ABC):
    def __init__(self):
        self._parent = None
        self._name = ''

    @property
    def parent(self) -> 'Element':
        return self._parent

    @parent.setter
    def parent(self, parent: 'Element'):
        self._parent = parent

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @abstractmethod
    def add(self, component: 'Element') -> 'Element':
        pass

    @abstractmethod
    def accept(self, visitor):
        pass


class PrimitiveElement(Element):
    def __init__(self, name='', values=''):
        super().__init__()
        self._name = name
        self._values = values

    def __repr__(self) -> str:
        return f'{self.name}: {self.values}' if self.name != '' else ''

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, values):
        self._values = values

    def add(self, component: 'Element') -> 'Element':
        pass

    def accept(self, visitor):
        return visitor.visit(self.name, self.values)


class ComplexElement(Element):
    def __init__(self, name='') -> None:
        super().__init__()
        self._children = []
        self._name = name

    def __repr__(self) -> str:
        children_repr = [f'{self.name}.{self.__ident(str(child))}' for child in self._children if str(child) != '']
        return '\n'.join(children_repr)

    def __ident(self, representation: str) -> str:
        return representation.replace('\n', '\n{}'.format(' ' * (len(self.name) + 1)))

    def add(self, component: Element) -> 'Element':
        if component:
            self._children.append(component)
            component.parent = self
        return self

    def remove(self, component: Element) -> None:
        self._children.remove(component)
        component.parent = None

    def accept(self, visitor):
        values = [child.accept(visitor) for child in self._children]
        return visitor.visit(self.name, values)



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


import sys
from pathlib import Path

from cmake_tidy.utils.app_configuration.configuration import Configuration, ConfigurationError


class FormatConfiguration(Configuration):
    def __init__(self, arguments: dict):
        super().__init__(arguments)
        self.__input_data = self.__initialize_input(arguments)

    @property
    def input(self) -> str:
        return self.__input_data

    @property
    def inplace(self) -> str:
        return self._config.get(self._property_name()) is True

    @property
    def file(self) -> Path:
        return Path(self._config['input'])

    @property
    def command(self) -> str:
        return 'format'

    def __initialize_input(self, arguments) -> str:
        return self.__load_input_data(arguments)

    @staticmethod
    def __load_input_data(arguments) -> str:
        try:
            return Path(arguments['input']).read_text()
        except Exception:
            raise ConfigurationError('Error - incorrect \"input\" - please specify existing file to be formatted')




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'BEGIN_PARENTHESIS BRACKET_ARGUMENT_CONTENT BRACKET_ARGUMENT_END BRACKET_ARGUMENT_START COMMAND_INVOCATION_END COMMAND_INVOCATION_START END_PARENTHESIS LINE_COMMENT NEWLINES QUOTED_ARGUMENT_CONTENT QUOTED_ARGUMENT_END QUOTED_ARGUMENT_START SPACES UNQUOTED_ARGUMENTfile : file file_element\n                | file_element\n                | emptyfile_element : line_ending\n                        | spaces\n                        | command_invocationcommand_invocation : start_cmd_invoke arguments end_cmd_invokeparentheses : start_parenthesis arguments end_parenthesisarguments : arguments argument\n                     | argumentargument : bracket_argument\n                    | quoted_argument\n                    | unquoted_argument\n                    | arguments_separation\n                    | parentheses\n                    | emptyarguments_separation : spaces\n                                | line_endingquoted_argument : QUOTED_ARGUMENT_START quoted_argument_content QUOTED_ARGUMENT_ENDbracket_argument : BRACKET_ARGUMENT_START bracket_argument_content BRACKET_ARGUMENT_ENDline_ending : line_comment newlines\n                       | newlinesunquoted_argument : UNQUOTED_ARGUMENTline_comment : LINE_COMMENTnewlines : NEWLINESempty :bracket_argument_content : bracket_argument_content bracket_argument_content_element\n                                    | bracket_argument_content_elementquoted_argument_content : quoted_argument_content quoted_argument_content_element\n                                   | quoted_argument_content_element\n                                   | emptybracket_argument_content_element : BRACKET_ARGUMENT_CONTENTquoted_argument_content_element : QUOTED_ARGUMENT_CONTENTspaces : SPACESstart_parenthesis : BEGIN_PARENTHESISend_parenthesis : END_PARENTHESISstart_cmd_invoke : COMMAND_INVOCATION_STARTend_cmd_invoke : COMMAND_INVOCATION_END'
    
_lr_action_items = {'SPACES':([0,1,2,3,4,5,6,8,9,10,12,13,14,15,16,17,18,19,20,21,22,23,26,27,28,29,30,31,32,33,41,42,44,46,47,],[9,9,-2,-3,-4,-5,-6,-22,-34,9,-25,-37,-1,-21,9,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,9,-35,-7,-9,-38,9,-20,-19,-8,-36,]),'LINE_COMMENT':([0,1,2,3,4,5,6,8,9,10,12,13,14,15,16,17,18,19,20,21,22,23,26,27,28,29,30,31,32,33,41,42,44,46,47,],[11,11,-2,-3,-4,-5,-6,-22,-34,11,-25,-37,-1,-21,11,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,11,-35,-7,-9,-38,11,-20,-19,-8,-36,]),'NEWLINES':([0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,26,27,28,29,30,31,32,33,41,42,44,46,47,],[12,12,-2,-3,-4,-5,-6,12,-22,-34,12,-24,-25,-37,-1,-21,12,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,12,-35,-7,-9,-38,12,-20,-19,-8,-36,]),'COMMAND_INVOCATION_START':([0,1,2,3,4,5,6,8,9,12,14,15,31,33,],[13,13,-2,-3,-4,-5,-6,-22,-34,-25,-1,-21,-7,-38,]),'$end':([0,1,2,3,4,5,6,8,9,12,14,15,31,33,],[-26,0,-2,-3,-4,-5,-6,-22,-34,-25,-1,-21,-7,-38,]),'COMMAND_INVOCATION_END':([8,9,10,12,13,15,16,17,18,19,20,21,22,23,26,27,28,32,42,44,46,47,],[-22,-34,-26,-25,-37,-21,33,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,-9,-20,-19,-8,-36,]),'BRACKET_ARGUMENT_START':([8,9,10,12,13,15,16,17,18,19,20,21,22,23,26,27,28,29,30,32,41,42,44,46,47,],[-22,-34,24,-25,-37,-21,24,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,24,-35,-9,24,-20,-19,-8,-36,]),'QUOTED_ARGUMENT_START':([8,9,10,12,13,15,16,17,18,19,20,21,22,23,26,27,28,29,30,32,41,42,44,46,47,],[-22,-34,25,-25,-37,-21,25,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,25,-35,-9,25,-20,-19,-8,-36,]),'UNQUOTED_ARGUMENT':([8,9,10,12,13,15,16,17,18,19,20,21,22,23,26,27,28,29,30,32,41,42,44,46,47,],[-22,-34,26,-25,-37,-21,26,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,26,-35,-9,26,-20,-19,-8,-36,]),'BEGIN_PARENTHESIS':([8,9,10,12,13,15,16,17,18,19,20,21,22,23,26,27,28,29,30,32,41,42,44,46,47,],[-22,-34,30,-25,-37,-21,30,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,30,-35,-9,30,-20,-19,-8,-36,]),'END_PARENTHESIS':([8,9,12,15,17,18,19,20,21,22,23,26,27,28,29,30,32,41,42,44,46,47,],[-22,-34,-25,-21,-10,-11,-12,-13,-14,-15,-16,-23,-17,-18,-26,-35,-9,47,-20,-19,-8,-36,]),'BRACKET_ARGUMENT_CONTENT':([24,34,35,36,43,],[36,36,-28,-32,-27,]),'QUOTED_ARGUMENT_CONTENT':([25,37,38,39,40,45,],[40,40,-30,-31,-33,-29,]),'QUOTED_ARGUMENT_END':([25,37,38,39,40,45,],[-26,44,-30,-31,-33,-29,]),'BRACKET_ARGUMENT_END':([34,35,36,43,],[42,-28,-32,-27,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'file':([0,],[1,]),'file_element':([0,1,],[2,14,]),'empty':([0,10,16,25,29,41,],[3,23,23,39,23,23,]),'line_ending':([0,1,10,16,29,41,],[4,4,28,28,28,28,]),'spaces':([0,1,10,16,29,41,],[5,5,27,27,27,27,]),'command_invocation':([0,1,],[6,6,]),'line_comment':([0,1,10,16,29,41,],[7,7,7,7,7,7,]),'newlines':([0,1,7,10,16,29,41,],[8,8,15,8,8,8,8,]),'start_cmd_invoke':([0,1,],[10,10,]),'arguments':([10,29,],[16,41,]),'argument':([10,16,29,41,],[17,32,17,32,]),'bracket_argument':([10,16,29,41,],[18,18,18,18,]),'quoted_argument':([10,16,29,41,],[19,19,19,19,]),'unquoted_argument':([10,16,29,41,],[20,20,20,20,]),'arguments_separation':([10,16,29,41,],[21,21,21,21,]),'parentheses':([10,16,29,41,],[22,22,22,22,]),'start_parenthesis':([10,16,29,41,],[29,29,29,29,]),'end_cmd_invoke':([16,],[31,]),'bracket_argument_content':([24,],[34,]),'bracket_argument_content_element':([24,34,],[35,43,]),'quoted_argument_content':([25,],[37,]),'quoted_argument_content_element':([25,37,],[38,45,]),'end_parenthesis':([41,],[46,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> file","S'",1,None,None,None),
  ('file -> file file_element','file',2,'p_file','cmake_parser.py',25),
  ('file -> file_element','file',1,'p_file','cmake_parser.py',26),
  ('file -> empty','file',1,'p_file','cmake_parser.py',27),
  ('file_element -> line_ending','file_element',1,'p_file_element','cmake_parser.py',35),
  ('file_element -> spaces','file_element',1,'p_file_element','cmake_parser.py',36),
  ('file_element -> command_invocation','file_element',1,'p_file_element','cmake_parser.py',37),
  ('command_invocation -> start_cmd_invoke arguments end_cmd_invoke','command_invocation',3,'p_command_invocation','cmake_parser.py',42),
  ('parentheses -> start_parenthesis arguments end_parenthesis','parentheses',3,'p_parentheses','cmake_parser.py',49),
  ('arguments -> arguments argument','arguments',2,'p_arguments','cmake_parser.py',56),
  ('arguments -> argument','arguments',1,'p_arguments','cmake_parser.py',57),
  ('argument -> bracket_argument','argument',1,'p_argument','cmake_parser.py',65),
  ('argument -> quoted_argument','argument',1,'p_argument','cmake_parser.py',66),
  ('argument -> unquoted_argument','argument',1,'p_argument','cmake_parser.py',67),
  ('argument -> arguments_separation','argument',1,'p_argument','cmake_parser.py',68),
  ('argument -> parentheses','argument',1,'p_argument','cmake_parser.py',69),
  ('argument -> empty','argument',1,'p_argument','cmake_parser.py',70),
  ('arguments_separation -> spaces','arguments_separation',1,'p_arguments_separation','cmake_parser.py',75),
  ('arguments_separation -> line_ending','arguments_separation',1,'p_arguments_separation','cmake_parser.py',76),
  ('quoted_argument -> QUOTED_ARGUMENT_START quoted_argument_content QUOTED_ARGUMENT_END','quoted_argument',3,'p_quoted_argument','cmake_parser.py',81),
  ('bracket_argument -> BRACKET_ARGUMENT_START bracket_argument_content BRACKET_ARGUMENT_END','bracket_argument',3,'p_bracket_argument','cmake_parser.py',86),
  ('line_ending -> line_comment newlines','line_ending',2,'p_line_ending','cmake_parser.py',94),
  ('line_ending -> newlines','line_ending',1,'p_line_ending','cmake_parser.py',95),
  ('unquoted_argument -> UNQUOTED_ARGUMENT','unquoted_argument',1,'p_unquoted_argument','cmake_parser.py',102),
  ('line_comment -> LINE_COMMENT','line_comment',1,'p_line_comment','cmake_parser.py',107),
  ('newlines -> NEWLINES','newlines',1,'p_newlines','cmake_parser.py',112),
  ('empty -> <empty>','empty',0,'p_empty','cmake_parser.py',118),
  ('bracket_argument_content -> bracket_argument_content bracket_argument_content_element','bracket_argument_content',2,'p_bracket_argument_content','cmake_parser.py',123),
  ('bracket_argument_content -> bracket_argument_content_element','bracket_argument_content',1,'p_bracket_argument_content','cmake_parser.py',124),
  ('quoted_argument_content -> quoted_argument_content quoted_argument_content_element','quoted_argument_content',2,'p_quoted_argument_content','cmake_parser.py',129),
  ('quoted_argument_content -> quoted_argument_content_element','quoted_argument_content',1,'p_quoted_argument_content','cmake_parser.py',130),
  ('quoted_argument_content -> empty','quoted_argument_content',1,'p_quoted_argument_content','cmake_parser.py',131),
  ('bracket_argument_content_element -> BRACKET_ARGUMENT_CONTENT','bracket_argument_content_element',1,'p_bracket_argument_content_element','cmake_parser.py',136),
  ('quoted_argument_content_element -> QUOTED_ARGUMENT_CONTENT','quoted_argument_content_element',1,'p_quoted_argument_content_element','cmake_parser.py',141),
  ('spaces -> SPACES','spaces',1,'p_spaces','cmake_parser.py',146),
  ('start_parenthesis -> BEGIN_PARENTHESIS','start_parenthesis',1,'p_start_parenthesis','cmake_parser.py',151),
  ('end_parenthesis -> END_PARENTHESIS','end_parenthesis',1,'p_end_parenthesis','cmake_parser.py',156),
  ('start_cmd_invoke -> COMMAND_INVOCATION_START','start_cmd_invoke',1,'p_start_cmd_invoke','cmake_parser.py',161),
  ('end_cmd_invoke -> COMMAND_INVOCATION_END','end_cmd_invoke',1,'p_end_cmd_invoke','cmake_parser.py',166),
]



###############################################################################
# Copyright Maciej Patro (maciej.patro@gmail.com)
# MIT License
###############################################################################


from typing import List

from cmake_tidy.formatting.utils.format_newline import FormatNewline
from cmake_tidy.formatting.utils.invocation.invocation_formatter import InvocationFormatter
from cmake_tidy.formatting.utils.invocation.invocation_wrapper import InvocationWrapper
from cmake_tidy.formatting.utils.tokens import Tokens


class ConditionFormatter(InvocationFormatter):
    def __init__(self, state: dict, settings: dict):
        super().__init__(state, settings)

    def format(self, invocation: dict) -> str:
        invocation['arguments'] = self._prepare_arguments(invocation)
        invocation['arguments'] = self.__split_invocation_if_needed(invocation)
        return self._join_command_invocation(invocation)

    def __split_invocation_if_needed(self, invocation: dict) -> list:
        if not self._is_fitting_in_line(invocation):
            invocation = InvocationWrapper().wrap(invocation)
            return self.__split_invocation(invocation['arguments'])
        return invocation['arguments']

    def __split_invocation(self, args: List[str]) -> list:
        for i in range(len(args)):
            self.__update_state(args[i])
            self.__replace_token_with_newline_if_needed(args, i)
        return args

    def __replace_token_with_newline_if_needed(self, args: List[str], index: int):
        argument_diff = -1 if self._settings.get('condition_splitting_move_and_or_to_newline') else 1
        if (args[index] == 'OR' or args[index] == 'AND') and self.__is_spacing_token(args, index + argument_diff):
            args[index + argument_diff] = FormatNewline(self._state, self._settings)(1)

    def __update_state(self, token: str):
        if token == '(':
            self._state['indent'] += 1
        elif token == ')':
            self._state['indent'] -= 1

    @staticmethod
    def __is_spacing_token(args: List[str], index: int) -> bool:
        try:
            return Tokens.is_spacing_token(args[index])
        except IndexError:
            return False
